/*$Id: MarsGPU.h 366 2009-01-12 13:34:05Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __MARS_GPU_H__
#define __MARS_GPU_H__

#include <cutil.h>
#include <glib-2.0/glib.h>

#define IN
#define OUT

#define __ALLOC__
#define __TIMER__

/******
 Public: Mars Runtime Interfaces
 ***/
typedef struct
{
    //--------------------------------
    //public
    //--------------------------------
#define MAP_ONLY            0x01
#define MAP_GROUP           0x02
#define MAP_GROUP_REDUCE    0x03
    gchar workflow;

#define OUTPUT_TO_HOST      0x01
    gchar attr;

    guint32 rec_per_thread;

    //--------------------------------
    //private
    //--------------------------------

    //rec
    //index: x-keyoffset, y-keysize, z-valoffset, w-valsize
    gpointer _input_keys;
    gpointer _input_vals;
    int4*   _input_index;
    gsize   _input_num;
    gsize   _input_keys_size;
    gsize   _input_vals_size;
    gsize   _input_index_size; 
    gsize   _input_keys_buf_size;
    gsize   _input_vals_buf_size;
    gsize   _input_index_buf_size;
    gchar   _input_in_gpu;

    gpointer _inter_keys;
    gpointer _inter_vals;
    int4*   _inter_index;
    gsize   _inter_num;
    int2*   _inter_group_index;
    gsize   _inter_group_num;
    gsize   _inter_keys_size;
    gsize   _inter_vals_size;
    gsize   _inter_index_size;
    gsize   _inter_group_index_size;
    gchar   _inter_in_gpu;

    
    gpointer _output_keys;
    gpointer _output_vals;
    int4*   _output_index;
    gsize   _output_num;
    gsize   _output_keys_size;
    gsize   _output_vals_size;
    gsize   _output_index_size;
    gchar   _output_in_gpu;
    int2*   _output_group_index;
    gsize   _output_group_num;
    gsize   _output_group_index_size;
    
    //gpu info
    gsize _max_size_dmem;
    gint32 _max_num_reg_per_block;
    gsize _max_size_smem_per_block;
    gint32 _size_warp; 
    gint32 _max_thread_per_block;
    gint32 _ver;
} Spec_t;

extern "C"
{
//setup and teardown
Spec_t* MarsSpecNew();
void MarsFinish(Spec_t**);

//add record
void MarsAddInputRec(Spec_t*, gpointer, gsize, gpointer, gsize);

//main procedure
void MarsSchedule(Spec_t*);

//print functions
void MarsPrintGPUInfo(Spec_t*);
void MarsPrintRecInfo(Spec_t*, gchar);

typedef void (*ProcessFunc_t)(gpointer key, gsize keysize, gpointer val, gsize valsize, gsize recnum, gpointer arg);
typedef void (*ProcessGroupFunc_t)(gpointer key, gsize keysize, gpointer vals, gsize valnum, gpointer arg);
#define IN_REC		0x01
#define INTER_REC	0x02
#define OUT_REC		0x03
void MarsProcessRec(Spec_t*, gchar, ProcessFunc_t, gpointer arg);
void MarsProcessGroup(Spec_t*, gchar, ProcessGroupFunc_t, gpointer arg);

__device__ void emit_inter_count(gsize keySize,
                               gsize valSize,
                               guint32* keyHist,
                               guint32* valHist,
                               guint32* indexHist);

__device__ void emit_inter(
                         gpointer key,
                         gsize keySize,
                         gpointer val,
                         gsize valSize,
                         guint32* ps_keyoffset,
                         guint32* ps_valoffset,
                         guint32* ps_count,
                         int2* keyval_offset,
                         guint32* cur_index,
                         gpointer output_keys,
                         gpointer output_vals,
                         int4* output_index
                          );
}

/******
 Private: MarsUtils Internal Functions
 ***/
extern gsize g_dmem;
extern gsize g_hmem;

extern "C"
{
//thread functions
void cuda_sync();

//memory functions
gpointer cuda_malloc(IN gsize size);
gpointer cuda_realloc_host(gpointer buf, gsize oldsize, gsize newsize);
void cuda_free(IN gpointer* buf, IN gsize size);

void cuda_memcpy_h2d(OUT gpointer des, IN gpointer src, IN gsize size);
void cuda_memcpy_d2h(OUT gpointer des, IN gpointer src, IN gsize size);

gpointer cuda_malloc_host(IN gsize size);
void cuda_free_host(IN gpointer* buf, IN gsize size);

//timer
void cuda_start_timer(guint* timer);
void cuda_stop_timer(gchar* msg, guint* timer);

//building blocks
guint32 prefixSum(guint32* d_inArr, guint32* d_outArr, guint32 numRecords );

void saven_initialPrefixSum(unsigned int maxNumElements);
int sort_GPU (void * d_inputKeyArray, 
              int totalKeySize, 
              void * d_inputValArray, 
              int totalValueSize, 
              int4 * d_inputPointerArray, 
              int rLen, 
              void * d_outputKeyArray, 
              void * d_outputValArray, 
              int4 * d_outputPointerArray, 
              int2 ** h_outputKeyListRange
	      );
}//MarsUtils

#endif //__MARS_GPU_H__
