/*$Id: MarsCPU.h 375 2009-01-14 11:23:36Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __MARS_CPU_H__
#define __MARS_CPU_H__

#include <sys/time.h>
#include <sys/unistd.h>
#include <sys/stat.h>
#include <sched.h>
#include <sys/types.h>
#include <glib-2.0/glib.h>

typedef struct
{
    int x;
    int y;
    int z;
    int w;
} int4;

typedef struct
{
    int x;
    int y;
} int2;

typedef struct
{
    int x;
    int y;
    int z;
} int3;


typedef struct
{
    //--------------------------------
    //public
    //--------------------------------
#define MAP_ONLY            0x01
#define MAP_GROUP           0x02
#define MAP_GROUP_REDUCE    0x03
    gchar workflow;

#define OUTPUT_TO_HOST      0x01
    gchar attr;

    //--------------------------------
    //private
    //--------------------------------
    guint32  num_map_worker;
    guint32  num_reduce_worker;

    //rec
    //index: x-keyoffset, y-keysize, z-valoffset, w-valsize
    gpointer _input_keys;
    gpointer _input_vals;
    int4*   _input_index;
    gsize   _input_num;
    gsize   _input_keys_size;
    gsize   _input_vals_size;
    gsize   _input_index_size; 
    gsize   _input_keys_buf_size;
    gsize   _input_vals_buf_size;
    gsize   _input_index_buf_size;

    gpointer _inter_keys;
    gpointer _inter_vals;
    int4*   _inter_index;
    gsize   _inter_num;
    gsize   _inter_group_num;
    int2*   _inter_group_index;
    gsize   _inter_keys_size;
    gsize   _inter_vals_size;
    gsize   _inter_index_size;
    gsize   _inter_group_index_size;

    
    gpointer _output_keys;
    gpointer _output_vals;
    gpointer _output_num;
    int4*   _output_index;
    gsize   _output_keys_size;
    gsize   _output_vals_size;
    gsize   _output_index_size;
    gsize   _output_group_index_size;
    int2*   _output_group_index;
    gsize   _output_group_num;
    
    //cpu info
    gsize _proc_num;
} Spec_t;

typedef struct
{
   gpointer input_keys;
   gpointer input_vals;
   int4*    input_index; 
   int2*    input_group_index; 
   gsize    num_thread;
   gsize    start_pos;
   gsize    end_pos;
   guint32  rec_per_thread;
   guint32  index;

   guint32* key_hist;
   guint32* val_hist;
   guint32* index_hist;
} MapCount_t;

typedef struct
{
   gpointer input_keys;
   gpointer input_vals;
   int4*    input_index; 
   int2*    input_group_index; 
   gsize    num_thread;
   gsize    start_pos;
   gsize    end_pos;
   guint32  rec_per_thread;
   guint32  index;

   gpointer output_keys;
   gpointer output_vals;
   int4*    output_index; 
   
   guint32* ps_keyoffset;
   guint32* ps_valoffset;
   guint32* ps_count;

   int2* keyval_offset;
   guint32* cur_index;
} Map_t;
/******
 Mars Runtime
 ***/
//setup and teardown
Spec_t* MarsSpecNew();
void MarsFinish(Spec_t**);

//add record
void MarsAddInputRec(Spec_t*, gpointer, gsize, gpointer, gsize);

//main procedure
void MarsSchedule(Spec_t*);
void MarsPrintRecInfo(Spec_t*, gchar);

typedef void (*ProcessFunc_t)(gpointer key, gsize keysize, gpointer val, gsize valsize, gsize recnum, gpointer arg);
typedef void (*ProcessGroupFunc_t)(gpointer key, gsize keysize, gsize valnum, gpointer arg, Spec_t*, guint32);
#define IN_REC		0x01
#define INTER_REC	0x02
#define OUT_REC		0x03
void MarsProcessRec(Spec_t*, gchar, gsize max_num, ProcessFunc_t, gpointer arg);
void MarsProcessGroup(Spec_t*, gchar, ProcessGroupFunc_t, gpointer arg);
gpointer GetVal(Spec_t*, guint32, guint32);
gpointer GetReduceVal(gpointer, int4*, int2*, guint32, guint32);
/******
 MarsUtils
 ***/
typedef struct timeval Timer_t;
void cpu_start_timer(Timer_t*);
void cpu_stop_timer(gchar*, Timer_t*);

#endif //__MARS_CPU_H__
