/*$Id $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __MAP_CU__
#define __MAP_CU__

__device__ void MAP_COUNT_FUNC//(void* key, void* val, size_t keySize, size_t valSize)
{
    EMIT_INTER_COUNT(sizeof(gfloat), sizeof(gfloat)); 
}

__device__ gfloat cnd(gfloat d)
{
    const gfloat       A1 = 0.31938153f;
    const gfloat       A2 = -0.356563782f;
    const gfloat       A3 = 1.781477937f;
    const gfloat       A4 = -1.821255978f;
    const gfloat       A5 = 1.330274429f;
    const gfloat RSQRT2PI = 0.39894228040143267793994605993438f;

    gfloat K = 1.0f / (1.0f + 0.2316419f * fabsf(d));

    gfloat cnd = RSQRT2PI * expf(- 0.5f * d * d) * 
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0f - cnd;

    return cnd;
}

__device__ void MAP_FUNC//(void* key, void* val, size_t keySize, size_t valSize)
{
    guint32 offset = *(guint32*)key;
    gfloat* buf = *(gfloat**)val + offset;

    gfloat sqrtT, expRT;
    gfloat d1, d2, CNDD1, CNDD2;

    sqrtT = sqrtf(buf[3]);
    d1 = (logf(buf[1] / buf[2]) + (buf[4] + 0.5f * buf[5] * buf[5]) * buf[3]) / (buf[5] * sqrtT);
    d2 = d1 -  buf[5] * sqrtT;

    CNDD1 = cnd(d1);
    CNDD2 = cnd(d2);

    expRT = expf(-  buf[4] *  buf[3]);
    gfloat call =  buf[1]* CNDD1 -  buf[2] * expRT * CNDD2;
    gfloat put  =  buf[2] * expRT * (1.0f - CNDD2) -  buf[1]* (1.0f - CNDD1);

    EMIT_INTER(&call, sizeof(gfloat), &put, sizeof(gfloat));
}

#endif // __MAP_CU__
