#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <stddef.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <ctype.h>
#include <time.h>

#include "MapReduceScheduler.h"
#define CHECK_ERROR

typedef struct
{
    gfloat* buf;
    gsize offset;
    gsize num;
} map_input_t;

typedef struct
{
    gfloat call;
    gfloat put;
} map_output_t;

gsize num_option;
gsize offset = 0;
gsize num = 2;

int cmp(void* arg1, void* arg2)
{
    return 0;
}

int bs_splitter(void *data_in, int req_units, map_args_t *out)
{
	map_input_t *input = (map_input_t*)data_in;

        input->offset = offset;
	out->data = input;
	out->length = sizeof(map_input_t);	
        offset += num;

	if (num + offset >= num_option)
        {
            num = num_option - offset;
 
            return 0;
        }
        else
            return 1;
}

gfloat cnd(gfloat d)
{
    const gfloat       A1 = 0.31938153f;
    const gfloat       A2 = -0.356563782f;
    const gfloat       A3 = 1.781477937f;
    const gfloat       A4 = -1.821255978f;
    const gfloat       A5 = 1.330274429f;
    const gfloat RSQRT2PI = 0.39894228040143267793994605993438f;

    gfloat K = 1.0f / (1.0f + 0.2316419f * fabsf(d));

    gfloat cnd = RSQRT2PI * expf(- 0.5f * d * d) * 
        (K * (A1 + K * (A2 + K * (A3 + K * (A4 + K * A5)))));

    if(d > 0)
        cnd = 1.0f - cnd;

    return cnd;
}

void bs_map(map_args_t *args)
{
    map_input_t *input = (map_input_t*)args->data;
	
    guint32 offset = input->offset*6;
    gfloat* buf = (gfloat*)input->buf + offset;

    gfloat sqrtT, expRT;
    gfloat d1, d2, CNDD1, CNDD2;

    sqrtT = sqrtf(buf[3]);
    d1 = (logf(buf[1] / buf[2]) + (buf[4] + 0.5f * buf[5] * buf[5]) * buf[3]) / (buf[5] * sqrtT);
    d2 = d1 -  buf[5] * sqrtT;

    CNDD1 = cnd(d1);
    CNDD2 = cnd(d2);

    expRT = expf(-  buf[4] *  buf[3]);
    gfloat call =  buf[1]* CNDD1 -  buf[2] * expRT * CNDD2;
    gfloat put  =  buf[2] * expRT * (1.0f - CNDD2) -  buf[1]* (1.0f - CNDD1);

    //printf("%f, %f\n", call, put);
    map_output_t output;
    output.call = call;
    output.put = put;
    emit_intermediate((void*)&output, (void *)&input->offset, sizeof(map_output_t));
}

int main(int argc, char *argv[]) 
{
	if (argc != 2)
	{
		printf("usage: %s filename\n", argv[0]);
		exit(-1);
	}	
Timer_t  timer;
	cpu_start_timer(&timer);
	
        int fd;
        struct stat finfo;
        char* fname = argv[1];
        char* fdata;
        CHECK_ERROR((fd = open(fname, O_RDONLY)) < 0);
        CHECK_ERROR(fstat(fd, &finfo) < 0);
        CHECK_ERROR((fdata = mmap(0, finfo.st_size + 1, 
           PROT_READ | PROT_WRITE, MAP_PRIVATE, fd, 0)) == NULL);
 
        float* data = (char*)fdata;

        num_option = finfo.st_size / sizeof(float) / 6;
        g_debug("num_option:%d", num_option);
	//InitMapReduce();

	map_input_t map_input;
        memset(&map_input, 0, sizeof(map_input_t));
	map_input.buf = data;			

	final_data_t result;

//	 Setup scheduler args
	scheduler_args_t sched_args;
	memset(&sched_args, 0, sizeof(scheduler_args_t));
	sched_args.task_data = &map_input;
	sched_args.map = bs_map;
	sched_args.reduce = NULL;
	sched_args.splitter = bs_splitter;
	sched_args.key_cmp = cmp;
	sched_args.unit_size = sizeof(float)*6;
	sched_args.partition = NULL; // use default
	sched_args.result = &result;
	sched_args.data_size = sizeof(map_input_t);
	sched_args.L1_cache_size = 1024 * 64;
	sched_args.num_map_threads = 4;
	sched_args.num_reduce_threads = 4;
	sched_args.num_merge_threads =  8;
	sched_args.num_procs = 4;
	sched_args.key_match_factor = 4;
	
  
	CHECK_ERROR(map_reduce_scheduler(&sched_args) < 0);
	cpu_stop_timer("all", &timer);
//	FinishMapReduce();

   CHECK_ERROR(munmap(fdata, finfo.st_size + 1) < 0);
   CHECK_ERROR(close(fd) < 0);
   return 0;
}
