/*$Id: MarsUtils.c 340 2009-01-09 03:08:11Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsCPU.h"
#include "compare.c"

void cpu_start_timer(Timer_t* timer)
{
    gettimeofday(timer, 0);
}

void cpu_stop_timer(gchar* msg, Timer_t * timer)
{
    Timer_t end;
    gettimeofday(&end, 0);

    g_print("%s - %f (ms)\n", msg, (gdouble)(end.tv_sec - timer->tv_sec)*1000+ (gdouble)(end.tv_usec-timer->tv_usec)/1000);
}

/*******************************************
 sort
 *******************************************/
typedef struct
{
    int4* index;
    gsize rec_num;
    guint32 start_pos;
} qsort_t;

typedef struct
{
    int4* input1;
    int4* input2;

    gsize rec_num1;
    gsize rec_num2;

    guint32 start_pos;
    int4* output;
} msort_t;

gpointer *g_keybuf;

static int cmp_wrapper(gpointer index1, gpointer index2)
{
    int4* i1 = (int4*)index1;
    int4* i2 = (int4*)index2;

    gpointer *key1 = (gchar*)g_keybuf + i1->x;
    gpointer *key2 = (gchar*)g_keybuf + i2->x;

    return compare(key1, i1->y, key2, i2->y);
}

void* quicksort(void* arg)
{
    qsort_t* q = (qsort_t*)arg;

    qsort((void*)(q->index + q->start_pos),
           q->rec_num, sizeof(int4), cmp_wrapper);
    return 0;
}

void* mergesort(void* arg)
{
   msort_t* m = (msort_t*)arg;
   //g_debug("start_pos:%d, num1:%d, num2:%d", m->start_pos, 
        //m->rec_num1, m->rec_num2);
   guint32 output_pos = m->start_pos;
   guint32 cur1 = 0;
   guint32 cur2 = 0;

   while (1)
   {
       if (cmp_wrapper(&m->input1[cur1], &m->input2[cur2]) < 0)
       {
           m->output[output_pos] = m->input1[cur1];
           ++cur1;
       }
       else
       {
           m->output[output_pos] = m->input2[cur2];
           ++cur2;
       }
       ++output_pos;

       if (cur1 < m->rec_num1 &&
           cur2 < m->rec_num2)
       {
           continue;
       }
       else if (cur1 >= m->rec_num1 &&
           cur2 < m->rec_num2)
       {
           memcpy((void*)(m->output+output_pos), 
                  (void*)(m->input2+cur2),
                  (m->rec_num2-cur2)*sizeof(int4));
           //g_debug("%d", m->rec_num2-cur2);
           break;
       }
       else if (cur1 < m->rec_num1 &&
                cur2 >= m->rec_num2)
       {
           memcpy((void*)(m->output+output_pos), 
                  (void*)(m->input1+cur1),
                  (m->rec_num1-cur1)*sizeof(int4));
           //g_debug("%d", m->rec_num2-cur1);
           break;
       }
       else if (cur1 >= m->rec_num1 &&
                cur2 >= m->rec_num2)
       {
           break;
       }
   }
   //g_debug("start_pos:%d, num1:%d, num2:%d", output_pos, 
        //cur1, cur2);
   return 0;
}

void SortGroup(Spec_t* spec)
{

    //parallel quick sort
    int i;
    guint32 thread_num = spec->_proc_num * 2;
    //guint32 thread_num = 1;
    qsort_t *q = (qsort_t*)g_malloc(sizeof(qsort_t)*thread_num);
    gsize rec_per_thread = spec->_inter_num / thread_num + 
                        (gsize)(spec->_inter_num%thread_num!=0);
    GThread **threads = (GThread**)g_malloc0(sizeof(GThread*)*thread_num);
    g_keybuf = spec->_inter_keys;

    for (i = 0; i < thread_num; i++)
    {
        q[i].index = spec->_inter_index;
        q[i].rec_num = rec_per_thread;
        q[i].start_pos = rec_per_thread*i;
        if (i == thread_num - 1)
            q[i].rec_num = spec->_inter_num - q[i].start_pos;

        threads[i] = g_thread_create(quicksort, &q[i], TRUE, NULL);    
    }
    for (i = 0; i < thread_num; i++)
        g_thread_join(threads[i]);

    //merge
    guint32 merge_thread = thread_num / 2;
    guint32 factor = 0;
    int4* buf = (int4*)g_malloc(sizeof(int4)*spec->_inter_num);
    int4* obuf = spec->_inter_index;
    int4* ibuf = buf;

    while (merge_thread > 0)
    {
        int4* tmp = obuf;
        obuf = ibuf;
        ibuf = tmp;

        //g_debug("merge_thread:%d", merge_thread);
        guint32 pow2 = (guint32)pow(2.0, (double)factor);
        msort_t* m = (msort_t*)g_malloc0(sizeof(msort_t)*merge_thread);
        guint32 offset = 0;
        for (i = 0; i < merge_thread; i++)
        {
            m[i].output = obuf;
            guint j;
            //g_debug("thread:%d", i);
            m[i].input1 = ibuf + q[offset].start_pos;
            m[i].start_pos = q[offset].start_pos;
            for (j = 0; j < pow2; j++)
                m[i].rec_num1 += q[offset+j].rec_num;
            offset += pow2;

            m[i].input2 = ibuf + q[offset].start_pos;
            for (j = 0; j < pow2; j++)
                m[i].rec_num2 += q[offset+j].rec_num;
            offset += pow2;

            threads[i] = g_thread_create(mergesort, &m[i], TRUE, NULL);   
//            mergesort(&m[i]);
       }

       for (i = 0; i < merge_thread; i++)
            g_thread_join(threads[i]);

        ++factor;
        merge_thread /= 2;
        if (merge_thread == 0)
        {
            g_free(spec->_inter_index);
            spec->_inter_index = m[0].output;
        }
        g_free(m);
    }    

    //group
    spec->_inter_group_index = (int2*)g_malloc0(spec->_inter_num*sizeof(int2));    
    
    spec->_inter_group_num = 1;
    spec->_inter_group_index[0].x = 0;
    int4 *pre_index = spec->_inter_index;
 
    for (i = 1; i < spec->_inter_num; i++)
    {
        if (cmp_wrapper(pre_index, &spec->_inter_index[i]) != 0)
        {
            spec->_inter_group_index[spec->_inter_group_num].x = i;
            spec->_inter_group_index[spec->_inter_group_num-1].y = i;
            spec->_inter_group_num++;
            pre_index = &(spec->_inter_index[i]);
        }
    }
    spec->_inter_group_index[spec->_inter_group_num-1].y = i;

    g_free(threads);
    g_free(q);
}
