/*$Id: main.c 357 2009-01-12 04:19:17Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsCPU.h"
#include "global.h"

guint32 col_num;
guint32 row_num;

void make_rec(gpointer key, gsize keysize, gpointer val, gsize valsize, gsize recnum, gpointer arg)
{
    int2* pos = (int2*)key;
    guint32* co = (guint32*)val;
    g_debug("row1:%d, row2:%d, covariance:%d", pos->x, pos->y, *co);
}

int main(int argc, char** argv)
{
    if (argc != 2)
    {
        g_print("usage: %s filename\n", argv[0]);
        exit(0);
    }

    Timer_t alltimer;
    cpu_start_timer(&alltimer);
    gchar* filename = argv[1];
    
    //get file content to main memory
    gchar* content = NULL;
    gsize  length = 0;
  
    Spec_t* spec = MarsSpecNew();
    spec->workflow = MAP_ONLY;
    spec->num_map_worker = 16;

    Timer_t timer;
    cpu_start_timer(&timer);
    if (!g_file_get_contents(filename, &content, &length, NULL))
        g_error("failed to get file content!");
    cpu_stop_timer("i/o", &timer); 
    
    cpu_start_timer(&timer);
    col_num = (guint32)sqrt((double)(length/sizeof(guint32)));
    row_num = col_num;
    g_debug("col_num:%d, row_num:%d", col_num, row_num);

   //first mapreduce
   guint32 i, j;
   info_t info;
   info.col_num = col_num;
   for (i = 0; i < row_num; i++)
   {
        info.row_id = i;
        MarsAddInputRec(spec,
                        &content, 
                        sizeof(gpointer),
                        &info,
                        sizeof(info_t)); 
    }
    cpu_stop_timer("preprocess", &timer); 

    MarsSchedule(spec);

    //second mapreduce
    Spec_t* spec2 = MarsSpecNew();
    spec2->workflow = MAP_ONLY;
    spec2->num_map_worker = 16;

   
    guint32* means = (guint32*)spec->_output_vals;
    info2_t info2;
    info2.col_num = col_num;
    cpu_start_timer(&timer);
    guint32 recnum = 0;
    for (i = 0; i < row_num-1; i++)
    {
        info2.row1_id = i;
        info2.row1_mean = means[i];
        for (j = i+1; j < row_num; j++)
        {
            recnum ++;
            info2.row2_id = j;
            info2.row2_mean = means[j];
            MarsAddInputRec(spec2,
                        &info2,
                        sizeof(info2_t),
                        &content, 
                        sizeof(gpointer)); 
        }
    }
    cpu_stop_timer("preprocess", &timer); 

    MarsSchedule(spec2);

//    MarsProcessRec(spec2, OUT_REC, recnum, make_rec, NULL);

    //clean
    g_free(content);
    MarsFinish(&spec);
    MarsFinish(&spec2);
    cpu_stop_timer("all", &alltimer);

    return 0;
}
