/*$Id: main.cu 419 2009-02-20 07:06:47Z wenbin $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MarsGPU.h"
#include "global.h"

guint32 col_num;
guint32 row_num;

int main(int argc, char** argv)
{
    if (argc != 2)
    {
        g_print("usage: %s filename\n", argv[0]);
        exit(0);
    }

    guint32 alltimer;
    cuda_start_timer(&alltimer);
    gchar* filename = argv[1];
    
    //get file content to main memory
    gchar* content = NULL;
    gsize  length = 0;
  
    Spec_t* spec = MarsSpecNew();
    spec->workflow = MAP_ONLY;
  //  spec->attr |= OUTPUT_TO_HOST;

    guint32 timer;
    cuda_start_timer(&timer);
    if (!g_file_get_contents(filename, &content, &length, NULL))
        g_error("failed to get file content!");
    cuda_stop_timer("i/o", &timer); 
    
    cuda_start_timer(&timer);
    col_num = (guint32)sqrt((double)(length/sizeof(guint32)));
    row_num = col_num;
    g_debug("col_num:%d, row_num:%d", col_num, row_num);

    gpointer d_content = cuda_malloc(length);;
    cuda_memcpy_h2d(d_content, content, length);
   //first mapreduce
   guint32 i, j;
   info_t info;
   info.col_num = col_num;
   for (i = 0; i < row_num; i++)
   {
        info.row_id = i;
        MarsAddInputRec(spec,
                        &d_content, 
                        sizeof(gpointer),
                        &info,
                        sizeof(info_t)); 
    }
    cuda_stop_timer("preprocess", &timer); 

    MarsSchedule(spec);

    //MarsProcessRec(spec, OUT_REC, print_rec, NULL);

    //second mapreduce
    Spec_t* spec2 = MarsSpecNew();
    spec2->workflow = MAP_ONLY;

   
    info2_t info2;
    info2.col_num = col_num;
    info2.means = (guint64)spec->_output_vals;
    cuda_start_timer(&timer);
    guint32 recnum = 0;
    for (i = 0; i < row_num-1; i++)
    {
        info2.row1_id = i;
        for (j = i+1; j < row_num; j++)
        {
            recnum ++;
            info2.row2_id = j;
            MarsAddInputRec(spec2,
                        &info2,
                        sizeof(info2_t),
                        &d_content, 
                        sizeof(gpointer)); 
        }
    }
    cuda_stop_timer("preprocess", &timer); 

    MarsSchedule(spec2);

    //clean
    g_free(content);
    MarsFinish(&spec);
    MarsFinish(&spec2);
    cuda_free(&d_content, length);
    cuda_stop_timer("all", &alltimer);

    return 0;
}
