/*$Id $*/

/*
# Copyright (c) 2008, Wenbin Fang (wenbin@cse.ust.hk), HKUST, Hong Kong
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the HKUST nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY  ''AS IS'' AND ANY
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Wenbin Fang BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef __MAP_CU__
#define __MAP_CU__

#include "global.h"

__device__ void MAP_COUNT_FUNC//(void* key, void* val, size_t keySize, size_t valSize)
{
    if (keySize == sizeof(gpointer))
    {
        EMIT_INTER_COUNT(sizeof(guint32), sizeof(guint32));
    }
    else
        EMIT_INTER_COUNT(sizeof(guint32)*2, sizeof(guint32));

}

__device__ void MAP_FUNC//(void* key, void* val, size_t keySize, size_t valSize)
{
    if (keySize == sizeof(gpointer))
    {
        guint32* matrix = *(guint32**)key;
        info_t* info = (info_t*)val;
        guint32* row = matrix + info->row_id * info->col_num;

        guint mean = 0;
        guint i;
        for (i = 0; i < info->col_num; i++)
            mean += row[i];
        mean /= info->col_num;

        EMIT_INTER(&info->row_id, sizeof(guint32), &mean, sizeof(guint32));
    }
    else
    {
        info2_t* info = (info2_t*)key;
        guint32* matrix = *(guint32**)val;
        guint32* row1 = (guint32*)matrix + info->row1_id * info->col_num;
        guint32* row2 = (guint32*)matrix + info->row2_id * info->col_num;
        guint32* means = (guint32*)info->means;
        
        guint32 covariance = 0;
        guint32 i;
        guint32 mean1 = means[info->row1_id];
        guint32 mean2 = means[info->row2_id];
        for (i = 0; i < info->col_num; i++)
            covariance += (row1[i]-mean1) * (row2[i]-mean2);
        covariance /= info->col_num;


        guint32 pos = info->row1_id * info->col_num + info->row2_id;

        EMIT_INTER(&info, sizeof(guint32)*2, &covariance, sizeof(guint32));
    }

}

#endif // __MAP_CU__
